<?php

namespace NsFWK;

use Aws\S3\S3Client;
use Aws\S3\MultipartUploader;
use Aws\S3\Exception\S3Exception;
use Aws\Exception\MultipartUploadException;
use Aws\Credentials\Credentials;

class ClsFileManagerObjectStorage {
    const NORMAL_FILE_LIMIT = 5000; // 100 MB (recommended limit for normal upload, accepts up to 5 GB)
    const UPLOAD_PART_SIZE = 104857600; // 100 MB (should be between 5 MB - 5 GB)

    protected S3Client $_objS3Client;

    /**
     * Constructor to set the client object and the bucket to be used for subsequent functions.
     *
     * @param array $arrConfiguration Configuration array with keys:
     *              accessKey  => string  (credentials of object storage)
     *              secretKey  => string  (credentials of object storage)
     *              endpoint   => string  (optional, if defined)
     *              pathStyle  => bool    (true => use path style, false => use virtual host bucket) 
     *              region     => string  (AWS service region)
     *              bucket     => string  (bucket name)
     */
    public function __construct(array $arrConfiguration) {   
        if (!$this->validateConfiguration($arrConfiguration)) {
            throw new \InvalidArgumentException('Invalid S3 configuration');
        }

        $credentials = new Credentials($arrConfiguration['accessKey'], $arrConfiguration['secretKey']);

        $arrArgs = [
            'region'      => $arrConfiguration['region'] ?? 'us-east-1',
            'version'     => 'latest',
            'credentials' => $credentials,
        ];

        if (!empty($arrConfiguration['endpoint'])) {
            $arrArgs['endpoint'] = $arrConfiguration['endpoint'];
        }

        if (!empty($arrConfiguration['pathStyle'])) {
            $arrArgs['use_path_style_endpoint'] = true;
        }

        $this->_objS3Client = new S3Client($arrArgs);
    }

    public function Read(string $strBucket, string $strKey): ?string {
        $arrParam = ['Bucket' => $strBucket, 'Key' => $strKey];
        try {
            $rslt = $this->_objS3Client->getObject($arrParam);
            return $rslt['Body']->getContents();
        } catch (S3Exception $e) {
            \NsFWK\ClsLogger::GetInstance()->LogDebug($e->getMessage());
            return null;
        }
    }

    public function Download(string $strBucket, string $strKey, string $strDownloadPath): bool {
        $arrParam = [
            'Bucket' => $strBucket,
            'Key' => $strKey,
            'SaveAs' => $strDownloadPath
        ];

        try {
            $rslt = $this->_objS3Client->getObject($arrParam);
            return $rslt['@metadata']['statusCode'] === 200;
        } catch (S3Exception $e) {
            \NsFWK\ClsLogger::GetInstance()->LogDebug($e->getMessage());
            return false;
        }
    }

    public function Delete(string $strBucket, string $strKey): bool {
        $arrParam = ['Bucket' => $strBucket, 'Key' => $strKey];
        try {
            $this->_objS3Client->deleteObject($arrParam);
            return true;
        } catch (S3Exception $e) {
            \NsFWK\ClsLogger::GetInstance()->LogDebug($e->getMessage());
            return false;
        }
    }

    public function Upload(string $strBucket, string $strKey, string $strFilePath): bool {
        $intFileSize = filesize($strFilePath);

        if ($intFileSize > self::NORMAL_FILE_LIMIT) {
            return $this->multiPartUpload($strBucket, $strKey, $strFilePath);
        } else {
            return $this->singlePartUpload($strBucket, $strKey, $strFilePath);
        }
    }

    protected function MultiPartUpload(string $strBucket, string $strKey, string $strFilePath): bool {
        $arrParam = [
            'Bucket' => $strBucket,
            'Key' => $strKey,
            'part_size' => self::UPLOAD_PART_SIZE,
            'ACL' => 'public-read'
        ];

        try {
            $uploader = new MultipartUploader($this->_objS3Client, $strFilePath, $arrParam);
            $rslt = $uploader->upload();
            return $rslt['@metadata']['statusCode'] === 200;
        } catch (MultipartUploadException $e) {
            $params = $e->getState()->getId();
            $this->_objS3Client->abortMultipartUpload($params);
            \NsFWK\ClsLogger::GetInstance()->LogDebug($e->getMessage());
            return false;
        }
    }

    protected function SinglePartUpload(string $strBucket, string $strKey, string $strFilePath): bool {
        $arrParam = [
            'Bucket' => $strBucket,
            'Key' => $strKey,
            'SourceFile' => $strFilePath,
            'ACL' => 'public-read'
        ];

        try {
            $rslt = $this->_objS3Client->putObject($arrParam);
            return $rslt['@metadata']['statusCode'] === 200;
        } catch (S3Exception $e) {
            \NsFWK\ClsLogger::GetInstance()->LogDebug($e->getMessage());
            return false;
        }
    }

    public function FileExists(string $strBucket, string $strKey): bool {
        try {
            return $this->_objS3Client->doesObjectExist($strBucket, $strKey);
        } catch (S3Exception $e) {
            \NsFWK\ClsLogger::GetInstance()->LogDebug($e->getMessage());
            return false;
        }
    }

    public function CreateFolder(string $strBucket, string $strFolder): bool {
        $arrData = [
            'Bucket' => $strBucket,
            'Key' => "$strFolder/",
            'Body' => "",
            'ACL' => 'public-read'
        ];

        try {
            $this->_objS3Client->putObject($arrData);
            return true;
        } catch (S3Exception $e) {
            \NsFWK\ClsLogger::GetInstance()->LogDebug($e->getMessage());
            return false;
        }
    }

    public function ListBuckets(): array {
        $arrBuckets = $this->_objS3Client->listBuckets();
        return $arrBuckets['Buckets'] ?? [];
    }

    public function ListObjectsInBucket(string $strBucket, string $strPrefix = ''): array {
        $arrParam = ['Bucket' => $strBucket];

        if (!empty($strPrefix)) {
            $arrParam['Prefix'] = $strPrefix;
        }

        $objectsListResponse = $this->_objS3Client->listObjects($arrParam);
        return $objectsListResponse['Contents'] ?? [];
    }

    private function validateConfiguration(array &$arrConfiguration): bool {
        return !empty($arrConfiguration['accessKey']) && !empty($arrConfiguration['secretKey']);
    }

    public function GetFileSize(string $strBucket, string $strKey): ?int {
        $arrParam = ['Bucket' => $strBucket, 'Key' => $strKey];
    
        try {
            $rslt = $this->_objS3Client->headObject($arrParam);
            return $rslt['ContentLength'] ?? null;
        } catch (S3Exception $e) {
            \NsFWK\ClsLogger::GetInstance()->LogDebug($e->getMessage());
            return null;
        }
    }
}
